package QEMU::Manager;

use base qw(Net::DBus::Object);

use Net::DBus::Exporter qw(com.redhat.qemu.Manager);

use QEMU::Manager::Machine;

use File::Spec::Functions qw(catfile catdir);
use File::Path;
use IO::File;

our $VERSION = "0.0.3";

sub new {
    my $class = shift;
    my $service = shift;
    my $self = $class->SUPER::new($service, "/qemu/manager");
    my $configDir = shift;

    $self->{machines} = {};
    $self->{configDir} = $configDir;

    bless $self, $class;
    
    $self->_load_machines();

    return $self;
}


sub _load_machines {
    my $self = shift;
    
    my $dir = $self->get_config("machines");
    opendir DIR, $dir
	or die "cannot read machines directory: $!";
    
    my @machines = map { s/\.conf$//; $_ } grep { /^[_a-z0-9]+\.conf$/i } readdir DIR;

    closedir DIR
	or die "cannot close directory: $!";
    
    foreach my $machine (@machines) {
	$self->_add_machine($machine);
    }
}

sub get_config {
    my $self = shift;
    my $type = shift;

    my $base = catdir($self->{configDir}, $type);
    mkpath([$base]);

    if (@_) {
	my $name = shift;
	my $ext = shift;
	return catfile($base, $name . $ext);
    } else {
	return $base;
    }
}

dbus_method("list_machines", [], [["array", "string"]]);
sub list_machines {
    my $self = shift;
    
    my @machines = keys %{$self->{machines}};
    return \@machines;
}

dbus_method("has_machine", ["string"], ["bool"]);
sub has_machine {
    my $self = shift;
    my $name = shift;
    
    return exists $self->{machines}->{$name} ? 1 : 0;
}

dbus_signal("machine_created", ["string"]);
dbus_method("create_machine", ["string"]);
sub create_machine {
    my $self = shift;
    my $name = shift;
    
    die "name must only consist of characters a-z, 0-9, _" 
	unless $name =~ /^[_a-z0-9]+$/i;
    
    die "a machine with name '$name' already exists"
	if $self->{machines}->{$name};

    my $config = $self->get_config("machines", $name, ".conf");
    my $fh = IO::File->new(">$config")
	or die "cannot create machine config $config: $!";
    
    $fh->close
	or die "cannot save machine config $config: $!";
    
    my $machine = $self->_add_machine($name);
    $self->emit_signal("machine_created", $name);
}

dbus_signal("machine_deleted", ["string"]);
dbus_method("delete_machine", ["string"]);
sub delete_machine {
    my $self = shift;
    my $name = shift;

    die "no machine with name '$name' currently exists"
	unless $self->{machines}->{$name};
    
    $self->_remove_machine($name);

    my $config = $self->get_config("machines", $name, ".conf");
    unlink $config
	or die "cannot remove machine config $config: $!";
    $self->emit_signal("machine_deleted", $name);
}

sub _add_machine {
    my $self = shift;
    my $name = shift;
    
    my $machine = QEMU::Manager::Machine->new($self, 
					      $name, 
					      $self->get_config("machines", $name, ".conf"),
					      $self->get_config("logs", $name, ".log"));
    
    $self->{machines}->{$name} = $machine;
}

sub _remove_machine {
    my $self = shift;
    my $name = shift;
    
    my $machine = $self->{machines}->{$name};
    $machine->disconnect;
    delete $self->{machines}->{$name};
}

1;
