package QEMU::Manager::Machine;


use strict;
use warnings;

use base qw(Net::DBus::Object);

use Net::DBus::Exporter qw(com.redhat.qemu.manager.Machine);

use IO::File;
use Config::Record;

use Net::DBus::Reactor;
use QEMU::Manager::Device::NIC;
use QEMU::Manager::Device::Display;
use QEMU::Manager::Device::Floppy;
use QEMU::Manager::Device::HardDrive;
use QEMU::Manager::Device::CDROM;
use QEMU::Manager::Device::SerialPort;
use QEMU::Manager::Device::ParallelPort;

use IPC::Open3;
use POSIX qw(F_GETFL F_SETFL  WNOHANG :errno_h strftime);
use Fcntl;
use Time::HiRes qw(sleep);

our %DEVICES = (
		nic => {
		    module => "QEMU::Manager::Device::NIC",
		    max => 100, # Arbitrary
		},
		display => {
		    module => "QEMU::Manager::Device::Display",
		    max => 1,
		},
		floppy => {
		    module => "QEMU::Manager::Device::Floppy",
		    max => 2,
		},
		harddrive => {
		    module => "QEMU::Manager::Device::HardDrive",
		    max => 4,
		},
		cdrom => {
		    module => "QEMU::Manager::Device::CDROM",
		    max => 1,
		},
		serialport => {
		    module => "QEMU::Manager::Device::SerialPort",
		    max => 4,
		},
		parallelport => {
		    module => "QEMU::Manager::Device::ParallelPort",
		    max => 1,
		},
		);

our %BOOT_DEVICES = (
		  "cdrom" => "d",
		  "harddrive" => "c",
		  "floppy" => "a",
		  # "network" => "?",
		  );



our $STATUS_STABLE = "stable";
our $STATUS_TESTING = "testing";
our $STATUS_EXPERIMENTAL = "experimental";

our %ARCHS = (
	      "x86" => {
		  binary => "qemu",
		  machines => [ "pc", "isapc" ],
	      },
	      "x86_64" => {
		  binary => "qemu-system-x86_64",
		  machines => [ "pc", "isapc" ],
	      },
	      "ppc" => {
		  binary => "qemu-system-ppc",
		  machines => [ "mac99", "g3bw", "prep" ],
	      },
	      "sparc" => {
		  binary => "qemu-system-sparc",
		  machines => [ "sun4m" ],
	      },
	      "sparc64" => {
		  binary => "qemu-system-sparc",
		  machines => [ "sun4u" ],
	      },
	      "mips" => {
		  binary => "qemu-system-mips",
		  machines => [ "mips" ],
	      },
	      "arm" => {
		  binary => "qemu-system-arm",
		  machines => [ "integratorcp" ],
	      },
	      );

our %MACHINES = (
		 "pc" => {
		     title => "x86/x86_64 PC",
		     status => $STATUS_STABLE,
		 },
		 "isapc" => {
		     title => "x86 PC (no-PCI)",
		     status => $STATUS_STABLE,
		 },
		 "mac99" => {
		     title => "Mac 99",
		     status => $STATUS_STABLE,
		 },
		 "g3bw" => {
		     title => "PowerMac G3",
		     status => $STATUS_STABLE,
		 },
		 "prep" => {
		     title => "Mac PREP",
		     status => $STATUS_STABLE,
		 },
		 "sun4m" => {
		     title => "JavaStation (Sun 4m)",
		     status => $STATUS_TESTING,
		 },
		 "sun4u" =>  {
		     title => "UltraSparce (Sun 4u)",
		     status => $STATUS_EXPERIMENTAL,
		 },
		 "mips" => {
		     title => "MIPS r4k",
		     status => $STATUS_EXPERIMENTAL,
		 },
		 "integratorcp" => {
		     title => "ARM Integrator/CP",
		     status => $STATUS_TESTING,
		 },
	      );


sub new {
    my $class = shift;
    my $manager = shift;
    my $name = shift;
    my $self = $class->SUPER::new($manager,
				  "/machine/" . $name);

    $self->{name} = $name;
    $self->{configFile} = shift;
    $self->{logFile} = shift;
    $self->{config} = Config::Record->new();
    
    bless $self, $class;
    
    if (-f $self->{configFile}) {
	$self->_load_config;
    }

    return $self;
}

dbus_method("get_name", [], ["string"]);
sub get_name {
    my $self = shift;
    return $self->{name};
}

# Try reasonably hard to make sure we don't leave QEMU processes spinning
# XXX could do better :-)
sub DESTROY {
    my $self = shift;
    if ($self->is_running) {
	$self->_kill;
    }
}

dbus_method("load_config");
dbus_signal("config_loaded");
sub _load_config {
    my $self = shift;

    foreach my $key (keys %{$self->{devices}}) {
	my $dev = $self->{devices}->{$key};
	$dev->disconnect();
    }
    $self->{devices} = {};

    $self->{config}->load($self->{configFile});
    $self->{serial} = $self->get_config_serial;

    $self->{devices} = {};
    foreach my $type (keys %{$self->{config}->get("devices", {})}) {
	foreach my $name (keys %{$self->{config}->get("devices/$type", {})}) {
	    $self->_load_device($type, $name);
	}
    }

    $self->{snapshots} = {};
    foreach my $name (keys %{$self->{config}->get("snapshots", {})}) {
	$self->_load_snapshot($name);
    }


    $self->emit_signal("config_loaded");    
}

sub _load_device {
    my $self = shift;
    my $type = shift;
    my $name = shift;

    my $mod = $DEVICES{$type}->{module};
    my $path = "/$type/$name";
	
    my $dev = $mod->new($self, $type, $name, $self->{config}->view("devices" . $path));
    $self->{devices}->{$path} = $dev;

    $dev->connect_to_signal("config_changed", sub {
	my $fieldname = shift;
	if ($type eq "cdrom" ||
	    $type eq "floppy") {
	    if ($fieldname eq "image") {
		if ($self->is_running) {
		    $self->change_disk($name, $dev->get_image);
		}
	    }
	}
	
	unless ($self->is_config_changed)  {
	    $self->{serial} = $self->get_config_serial + 1;
	}
    });


}


sub _load_snapshot {
    my $self = shift;
    my $name = shift;

    $self->{snapshots}->{$name} = QEMU::Manager::Snapshot->new($self, $name, 
							       $self->{config}->view("snapshots/$name"));
}

dbus_method("save_config");
dbus_signal("config_saved");
sub save_config {
    my $self = shift;
    
    return unless $self->is_config_changed;

    $self->{config}->save($self->{configFile});
    $self->emit_signal("config_saved");
}

sub get_config {
    my $self = shift;
    my $name = shift;
    my $default = shift;
    return $self->{config}->get($name, $default);
}


dbus_signal("config_changed", ["string", "string"]);
sub set_config {
    my $self = shift;
    my $name = shift;
    my $value = shift;
    my $caller = shift;
    $self->{config}->set($name, $value);
    unless ($self->is_config_changed)  {
	$self->{serial} = $self->get_config_serial + 1;
    }
    $self->emit_signal("config_changed", $name, $caller);
}

dbus_method("is_config_changed", [], ["bool"]);
sub is_config_changed {
    my $self = shift;
    
    return $self->{serial} > $self->get_config_serial();
}


dbus_method("get_config_serial", [], ["uint32"]);
sub get_config_serial {
    my $self = shift;
    
    return $self->get_config("serial", 0);
}

dbus_method("get_memory", [], ["uint32"]);
sub get_memory {
    my $self = shift;
    return $self->get_config("memory", 64);
}

dbus_method("set_memory", ["uint32", "caller"]);
sub set_memory {
    my $self = shift;
    my $max = shift;
    my $caller = shift;
    if ($max < 32) {
	die "maximum memory must be >= 32 MB";
    }
    # XXX Probably a bit lame
    if ($max > 4096) {
	die "maximum memory must be <= 4 GB";
    }
    $self->set_config("memory", $max, $caller);
}

dbus_method("get_title", [], ["string"]);
sub get_title {
    my $self = shift;
    return $self->get_config("title", "QEMU");
}

dbus_method("set_title", ["string", "caller"]);
sub set_title {
    my $self = shift;
    my $title = shift;
    my $caller = shift;
    if (length $title < 1) {
	die "maximum title must be >= 1";
    }
    # XXX Probably a bit lame
    if (length $title > 128) {
	die "maximum title must be <= 128 characters";
    }
    $self->set_config("title", $title, $caller);
}

dbus_method("get_log", [], ["string"]);
sub get_log {
    my $self = shift;
    return $self->{logFile};
}

dbus_method("get_cpus", [], ["uint32"]);
sub get_cpus {
    my $self = shift;
    return $self->get_config("cpus", 1);
}

dbus_method("set_cpus", ["uint32", "caller"]);
sub set_cpus {
    my $self = shift;
    my $max = shift;
    my $caller = shift;
    if ($max < 1) {
	die "maximum cpus must be >= 1";
    }
    # XXX Probably a bit lame
    if ($max > 32) {
	die "maximum cpus must be <= 32";
    }
    $self->set_config("cpus", $max, $caller);
}

dbus_method("get_architecture", [], ["string"]);
sub get_architecture {
    my $self = shift;
    return $self->get_config("architecture", "x86");
}

dbus_method("set_architecture", ["string", "caller"]);
sub set_architecture {
    my $self = shift;
    my $arch = shift;
    my $caller = shift;
    unless (exists $ARCHS{lc $arch}) {
	die "unsupported architecture '$arch'";
    }
    $self->set_config("architecture", lc $arch, $caller);
    $self->set_machine($ARCHS{lc $arch}->{machines}->[0], $caller);
}


dbus_method("get_machine", [], ["string"]);
sub get_machine {
    my $self = shift;
    return $self->get_config("machine", "pc");
}

dbus_method("set_machine", ["string", "caller"]);
sub set_machine {
    my $self = shift;
    my $arch = shift;
    my $caller = shift;
    unless (exists $MACHINES{lc $arch}) {
	die "unsupported machine '$arch'";
    }
    $self->set_config("machine", lc $arch, $caller);
}


dbus_method("list_machines", [], [["array", "string"]]);
sub list_machines {
    my $self = shift;
    my @archs = sort { $a cmp $b } keys %MACHINES;
    return \@archs;
}

dbus_method("list_architectures", [], [["array", "string"]]);
sub list_architectures {
    my $self = shift;
    my @archs = sort { $a cmp $b } keys %ARCHS;
    return \@archs;
}

dbus_method("list_arch_machines", ["string"], [["array", "string"]]);
sub list_arch_machines {
    my $self = shift;
    my $arch = shift;
    my @machines = sort { $a cmp $b } @{$ARCHS{$arch}->{machines}};
    return \@machines;
}

dbus_method("get_machine_info", ["string"], [["dict", "string", "string"]]);
sub get_machine_info {
    my $self = shift;
    my $machine = shift;
    my %info = %{$MACHINES{$machine}};
    return \%info;
}



dbus_method("get_boot_device", [], ["string"]);
sub get_boot_device {
    my $self = shift;
    return $self->get_config("boot_device", "harddrive");
}


dbus_method("set_boot_device", ["string", "caller"]);
sub set_boot_device {
    my $self = shift;
    my $device = shift;
    my $caller = shift;
    unless (exists $BOOT_DEVICES{lc $device}) {
	die "unknown boot device $device";
    }
    $self->set_config("boot_device", lc $device, $caller);
}


dbus_method("list_devices", [], [["array", "string"]]);
sub list_devices {
    my $self = shift;
    
    my @devs = keys %{$self->{devices}};
    return \@devs;
}


dbus_method("list_devices_by_type", ["string"], [["array", "string"]]);
sub list_devices_by_type {
    my $self = shift;
    my $type = shift;
    
    my @devs = grep { m,/$type/, } keys %{$self->{devices}};
    return \@devs;
}


dbus_method("create_device", ["string", "string"], ["string"]);
dbus_signal("device_created", ["string"]);
sub create_device {
    my $self = shift;
    my $type = shift;
    my $name = shift;
    
    if (!defined $self->{config}->get("devices", undef)) {
	$self->{config}->set("devices", {});
    }

    if (!defined $self->{config}->get("devices/$type", undef)) {
	$self->{config}->set("devices/$type", {});
    }
    
    my $path = "/$type/$name";
    my $dev = $self->{config}->get("devices" . $path, undef);
    if (defined $dev) {
	die "device '$path' already exists";
    }

    if (!exists $DEVICES{$type}) {
	die "unknown device type '$type'";
    }

    $self->{config}->set("devices" . $path, {});    
    $self->_load_device($type, $name);
    $self->emit_signal("device_created", $path);
    
    return "/device" . $path;
}


sub get_device {
    my $self = shift;
    my $type = shift;
    my $name = shift;
    my $path = "/$type/$name";
    return $self->{devices}->{$path};
}

dbus_method("delete_device", ["string","string"]);
dbus_signal("device_deleted", ["string"]);
sub delete_device {
    my $self = shift;
    my $type = shift;
    my $name = shift;
    
    my $path = "/$type/$name";
    if (!exists $self->{devices}->{$path}) {
	die "device '$path' does not exist";
    }
    
    $self->{devices}->{$path}->disconnect;
    delete $self->{devices}->{$path};
    
    $self->{config}->set("devices/$path", undef);
    
    $self->emit_signal("device_deleted", $path);
}


dbus_method("save_snapshot",["string"], ["string"]);
dbus_signal("snapshot_saved", ["string"]);
sub save_snapshot {
    my $self = shift;
    my $directory = shift;

    die "machine is not running"
	unless $self->is_running;

    my $now = time;
    my $name = strftime("%Y-%m-%d-%H-%M-%S", gmtime $now);
    my $file = catfile($directory, $self->get_name . "-" . $name . ".qvm");

    my ($got_prompt, $lines) = $self->send_monitor_command("savevm $file");
    unless ($got_prompt) {
	unlink $file;
	die "cannot stop execution";
    }
    
    $self->config->set("snapshots/$name", {
	filename => $file,
	date => $now,
	label => "Snapshot on " . $name,
	serial => $self->get_config_serial,
    });
    $self->save_config;

    my $ss = $self->_load_snapshot($name);
    
    $self->emit_signal("snapshot_saved", $name);
}

dbus_method("load_snapshot", ["string"]);
sub load_snapshot {
    my $self = shift;
    my $name = shift;
    
    die "snapshot with name '$name' does not exist"
	unless exists $self->{snapshots}->{$name};
    
    unless ($self->is_running) {
	# XXX this won't work yet because we need to 
	# recursively enter the main loop until
	# we get the monitor tty back
	#$self->start(0);
	die "machine not running";
    }

    my $ss = $self->{snapshots}->{$name};
    
    my $file = $ss->get_filename;
    
    my ($got_prompt, $lines) = $self->send_monitor_command("loadvm $file");
    die "cannot stop execution" 
	unless $got_prompt;

    $self->emit_signal("snapshot_loaded", $name);
}


dbus_method("stop_cpu");
dbus_signal("cpu_stopped");
sub stop_cpu {
    my $self = shift;

    die "machine is not running"
	unless $self->is_running;
    die "machine is already stopped"
	if $self->is_stopped;
    
    my ($got_prompt, $lines) = $self->send_monitor_command("stop");
    die "cannot stop execution" 
	unless $got_prompt;

    $self->{stopped} = 1;
    $self->emit_signal("cpu_stopped");
}


dbus_method("cont_cpu");
dbus_signal("cpu_continued");
sub cont_cpu {
    my $self = shift;
    
    die "machine is not running"
	unless $self->is_running;
    die "machine is not stopped"
	unless $self->is_stopped;

    my ($got_prompt, $lines) = $self->send_monitor_command("cont");
    die "cannot continue execution" 
	unless $got_prompt;

    $self->{stopped} = 0;
    $self->emit_signal("cpu_continued");
}

dbus_method("is_stopped", [], ["bool"]);
sub is_stopped {
    my $self = shift;
    
    return $self->is_running &&
	$self->{stopped};
}


dbus_method("start");
dbus_signal("machine_started");
sub start {
    my $self = shift;
    my $stopped = shift;

    die "machine is already active" 
	if $self->is_running;
    
    my @cmd = $self->get_command_line;
    my %env = $self->get_environment;
    
    if ($stopped) {
	push @cmd, "-S";
    }

    local %ENV = %ENV;
    foreach my $key (keys %env) {
	$ENV{$key} = $env{$key};
    }

    my ($stdin, $stdout, $stderr);
    $stdin = IO::Handle->new();
    $stdout = IO::Handle->new();
    $stderr = IO::Handle->new();
    
    my $pid = open3($stdin, $stdout, $stderr, @cmd);
    # QEMU expects commands via the monitor, not stdin
    $stdin->close;
    
    my $log = IO::File->new($self->get_log, "w")
	or die "cannot create logfile " . $self->get_log . ": $!";
    
    print $log "Starting qemu at ", strftime("%H:%M:%S on %e %B %Y", gmtime time), "\n";
    print $log join(" ", @cmd), "\n";
    if (%env) {
	print $log "With environment variables\n";
	foreach my $key (keys %env) {
	    print $log $key, " -> ", $env{$key}, "\n";
	}
    }
    print $log "\n";
    $log->flush;

    $self->{stopped} = $stopped;
    $self->{pid} = $pid;
    $self->{log} = $log;
    $self->{stdout} = $stdout;
    $self->{stderr} = $stderr;
    $self->_setup_handle_nonblock($stdout);
    $self->_setup_handle_nonblock($stderr);

    my $reactor = Net::DBus::Reactor->main;
    $reactor->add_read($stdout->fileno, Net::DBus::Callback->new(object => $self,
								 method => "_handle_log_data",
								 args => [ $stdout ]));
    $reactor->add_read($stderr->fileno, Net::DBus::Callback->new(object => $self,
								 method => "_handle_log_data",
								 args => [ $stderr, "_watch_for_tty" ]));
    

    $self->emit_signal("machine_started");
    if ($stopped) {
	$self->emit_signal("cpu_stopped");
    }
}

sub _setup_handle_nonblock {
    my $self = shift;
    my $handle = shift;
    
    my $flags = 0;
    fcntl($handle, F_GETFL, $flags)
        or die "Couldn't get flags for handle : $!";
    
    $flags |= O_NONBLOCK;
    fcntl($handle, F_SETFL, $flags)
        or die "Couldn't set flags for handle: $!";
}


dbus_method("shutdown");
sub shutdown {
    my $self = shift;
    
    die "machine is not active"
	unless $self->is_running;
    
    $self->_kill;
}

dbus_signal("machine_stopped");
sub _kill {
    my $self = shift;

    kill &POSIX::SIGTERM, $self->{pid};
    my $got = waitpid $self->{pid}, WNOHANG;
    if ($got != $self->{pid}) {
	kill &POSIX::SIGKILL, $self->{pid};
	sleep 1;
	$got = waitpid $self->{pid}, WNOHANG;
    }

    $self->{rtty}->close if $self->{rtty};
    delete $self->{rtty};

    $self->{wtty}->close if $self->{wtty};
    delete $self->{rtty};

    if ($self->{stdout}) {
	my $reactor = Net::DBus::Reactor->main;
        $reactor->remove_read($self->{stdout}->fileno);
	# Suck down all pending data & log it
	while ($self->_save_log_data($self->{stdout}, $self->{log}) == 1) {};
	$self->{stdout}->close;
	delete $self->{stdout};
    }

    if ($self->{stderr}) {
	my $reactor = Net::DBus::Reactor->main;
        $reactor->remove_read($self->{stderr}->fileno);
	# Suck down all pending data & log it
	while ($self->_save_log_data($self->{stderr}, $self->{log}) == 1) {};
	$self->{stderr}->close;
	delete $self->{stderr};
    }
    
    $self->{log}->close if $self->{log};
    delete $self->{log};
    delete $self->{pid};
    $self->emit_signal("machine_stopped");
}

dbus_method("is_running", [], ["bool"]);
sub is_running {
    my $self = shift;
    
    return $self->{pid} ? 1 : 0;
}

sub _watch_for_tty {
    my $self = shift;
    my $data = shift;
    
    if ($data =~ m,char device redirected to (/dev/pts/(\d+)),) {
	my $tty = $1;

	eval {
	    my $rtty = IO::File->new($tty, "r")
		or die "cannot open $tty for reading: $!";
	    my $wtty = IO::File->new($tty, "w")
		or die "cannot open $tty for writing: $!";
	    
	    $self->{rtty} = $rtty;
	    $self->{wtty} = $wtty;

	    $self->_setup_handle_nonblock($rtty);
	    $self->_setup_handle_nonblock($wtty);	    
	    
	    # Suck first monitor prompt out
	    my ($got_prompt, $lines) = $self->read_until_monitor_prompt;
	};
	if ($@) {
	    my $err = $@;
	    $self->_kill;
	    die $err;
	}
    }
}

sub _handle_log_data {
    my $self = shift;
    my $handle = shift;
    my $hook = shift;

    # During cleanup there is a chance a calback will be invoked
    # by the reactor even after we're all done, so this is an
    # important sanity check
    return unless $self->is_running;

    my $ret = $self->_save_log_data($handle, $self->{log}, $hook);
    if ($ret == 0) {
        $self->_kill($handle);
    }
}

sub _save_log_data {
    my $self = shift;
    my $src = shift;
    my $dst = shift;
    my $hook = shift;

    my $buffer = '';
    my $rv = $src->sysread($buffer, 4096);
    if (!defined($rv)) {
        if ($! == EAGAIN) {
            # would block
            return -1;
        } else {
            die "cannot read from pipe " . $src->fileno . ": $!";
        }
    } else {
        if ($rv) {
            # successfully read $rv bytes from HANDLE
	    $dst->syswrite($buffer, $rv);
	    $dst->flush;
	    if ($hook) {
		$self->$hook($buffer);
	    }
            return 1;
        } else {
            # End of file
            return 0;
        }
    }
}


sub send_monitor_command {
    my $self = shift;
    my $cmd = shift;
    
    die "monitor is only available if machine is running"
	unless $self->is_running;
    
    my $fh = $self->{wtty};
    print $fh $cmd, "\n";
    $fh->flush;

    sleep 0.05;
    return $self->read_until_monitor_prompt;
}

sub read_until_monitor_prompt {
    my $self = shift;
    my $iter = shift;
    $iter = 10 unless $iter;

    die "monitor is only available if machine is running"
	unless $self->is_running;

    my $mon = $self->{rtty};
    
    my @lines;
    my $got_prompt = 0;
    for (my $i = 0 ; $i < $iter ; $i++) {
	while (defined (my $line = <$mon>)) {
	    chomp $line;
	    push @lines, $line;
	    if ($line =~ /^\(qemu\)\s*$/) {
		$got_prompt = 1;
		last;
	    }
	}
	last if $got_prompt;
	sleep 0.05;
    }
    return ($got_prompt, \@lines);
}

sub change_disk {
    my $self = shift;
    my $name = shift;
    my $image = shift;
    
    my ($got_prompt, $lines) = $self->send_monitor_command("info block");
    
    ($got_prompt, $lines) = $self->send_monitor_command("eject $name");
    
    ($got_prompt, $lines) = $self->send_monitor_command("change $name $image");
        
    ($got_prompt, $lines) = $self->send_monitor_command("info block");
}

sub get_environment {
    my $self = shift;

    my %env;
    foreach my $name (sort { $a cmp $b } keys %{$self->{devices}}) {
	my $dev = $self->{devices}->{$name};
	
	my %devEnv = $dev->get_environment_variables();
	foreach my $key (keys %devEnv) {
	    $env{$key} = $devEnv{$key};
	}
    }
    
    return %env;
}

sub get_command_line {
    my $self = shift;
    
    my $has_display = 0;
    my $has_nic = 0;
    my $has_serial = 0;
    my $has_parallel = 0;

    my $arch = $self->get_architecture;
    my $binary = $ARCHS{$arch}->{binary};

    my @args = ($binary, "-monitor", "pty");
    foreach my $name (sort { $a cmp $b } keys %{$self->{devices}}) {
	my $dev = $self->{devices}->{$name};
	if ($dev->isa("QEMU::Manager::Device::Display")) {
	    $has_display = 1;
	}
	if ($dev->isa("QEMU::Manager::Device::NIC")) {
	    $has_nic = 1;
	}
	if ($dev->isa("QEMU::Manager::Device::SerialPort")) {
	    $has_serial = 1;
	}
	if ($dev->isa("QEMU::Manager::Device::ParallelPort")) {
	    $has_parallel = 1;
	}
	push @args, $dev->get_command_line_args;
    }
    unless ($has_display) {
	push @args, "-nographic";
    }
    unless ($has_nic) {
	push @args, "-net", "none";
    }
    unless ($has_serial) {
	push @args, "-serial", "null";
    }
    unless ($has_parallel) {
	push @args, "-parallel", "null";
    }
    push @args, "-M", $self->get_machine;
    push @args, "-boot", $BOOT_DEVICES{$self->get_boot_device};
    push @args, "-m", $self->get_memory;
    push @args, "-smp", $self->get_cpus;
    
    my $t = $self->get_title;
    $t =~ s/"/\\"/g;
    
    push @args, "-title", $t;
    return @args;
}

1;

